import logging
import os
from logging.config import dictConfig

SWB_METER_LOG_LEVEL = os.getenv("SWB_METER_LOG_LEVEL", "INFO").upper()

# ロギング設定
logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
    },
    "handlers": {
        "default": {
            "level": SWB_METER_LOG_LEVEL,
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # 標準出力へログを出力
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": SWB_METER_LOG_LEVEL,
            "propagate": True,
        }
    },
}

dictConfig(logging_config)

logger = logging.getLogger(__name__)



import asyncio
from asyncio import timeout as async_timeout
from datetime import datetime

import click
from bleak import BleakScanner

from swb_meter.db import get_meter_table, get_temperature_table
from swb_meter.logger import logger
from swb_meter.parser import parse_advertisement_data


class DataHandler:
    """A class that handles data received from SwitchBot devices.

    Attributes:
        t (TemperatureTable): The temperature table used for data insertion.

    Methods:
        on_data(advertisement_data): Processes the advertisement data and inserts it into the temperature table.

    """

    def __init__(self):
        self.t = get_temperature_table()

    def on_data(self, advertisement_data):
        parsed_data = parse_advertisement_data(advertisement_data)

        if not parsed_data:
            return False

        insert_dict = {
            "mac_address": parsed_data.manufacturer_data.mac_address,
            "temperature": parsed_data.service_data.temperature.value,
            "scale": parsed_data.service_data.temperature.scale,
            "humidity": parsed_data.service_data.humidity,
            "battery": parsed_data.service_data.battery,
            "rssi": int(
                parsed_data.rssi
            ),  # `objc._pythonify.OC_PythonLong`型でエラーが出るのでintに変換
            "tx_power": parsed_data.tx_power,
            "created_at": datetime.now().strftime(
                "%Y-%m-%d %H:%M:%S"
            ),  # タイムスタンプを追加
        }

        self.t.insert(insert_dict)
        logger.info("inserted data.")
        logger.debug(parsed_data)

        return True


@click.group()
@click.version_option()
def cli():
    """SwitchBot Meters Server CLI tool.

    `swb-meter` allows you to listen for SwitchBot meters and insert the data into a SQLite database.

    By default the backend database is configured current directory with the name `swb_meter.db` (created if not exists).

    To change the database path by setting the environment variable `SWB_METER_DB_PATH`:

    \b
        $ export SWB_METER_DB_PATH=/path/to/db.sqlite

    \b
        or:

    \b
        $ SWB_METER_DB_PATH=/path/to/db.sqlite swb-meter listen


    To get started, add a meter with an alias (typically the room name where the meter is located) with its MAC address:

    \b
        $ swb-meter add "Living Room" 12:34:56:78:90:AB

    This will add a record to the `Meter` table.

    \b
        $ swb-meter ls
        Living Room 12:34:56:78:90:AB
        Total: 1 meters

    Then, start listening for the meters:

    \b
        $ swb-meter listen

    The data will be inserted into the `Temperature` table in the database.


    To visualize the data, you can run the Streamlit app in another terminal.
    You need optional dependencies [streamlit] installed:

    \b
        $ pip install swb-meter[streamlit]

    then run the Streamlit app:

    \b
        $ swb-meter streamlit


    By default, the log level is set to `INFO`.

    To change log level, set environment variable `SWB_METER_LOG_LEVEL=DEBUG` for verbose output.

    \b
        $ export SWB_METER_LOG_LEVEL=DEBUG

        or:

    \b
        $ SWB_METER_LOG_LEVEL=DEBUG swb-meter listen
    """


@cli.command(name="add")
@click.argument("alias", type=str, required=True)
@click.argument("mac_address", type=str, required=True)
@click.option("-u", "--upsert", is_flag=True, help="Upsert the existing record's alias")
def add_meter(alias, mac_address, upsert):
    """Add a SwitchBot meter to the master table with an alias.

    Example:
    \b

        $ swb-meter add "Living Room" 12:34:56:78:90:AB
    """

    # validate mac_address format
    if not mac_address.count(":") == 5:
        click.echo("Invalid MAC address format")
        return

    m = get_meter_table()

    if upsert:
        m.upsert({"alias": alias, "mac_address": mac_address}, pk="mac_address")
        click.echo("Upserted a meter")
    else:
        m.insert({"alias": alias, "mac_address": mac_address}, pk="mac_address")
        click.echo("Added a meter")


@cli.command(name="ls")
def list_meters():
    """List all SwitchBot meters in the master table.

    Example:
    \b

        $ swb-meter ls
    """
    m = get_meter_table()

    if not m.exists():
        click.echo("No meters added yet.")
        return

    for row in m.rows:
        click.echo(row)

    click.echo(f"Total: {m.count} meters")


@cli.command(name="rm")
@click.argument("mac_address", type=str, required=True)
def remove_meter(mac_address):
    """Remove a SwitchBot meter from the master table.

    Example:
    \b

        $ swb-meter rm 12:34:56:78:90:AB
    """

    m = get_meter_table()
    m.delete(mac_address)
    click.echo("Removed a meter")


@cli.command(name="listen")
@click.option(
    "-t",
    "--timeout",
    type=float,
    default=10,
    help="Timeout in seconds for each scan",
)
@click.option(
    "-i",
    "--interval",
    default=60,
    type=int,
    help="Interval in seconds between scans",
)
@click.option(
    "-s",
    "--scans",
    default=None,
    type=int,
    help="Number of scans. Default is None (infinite)",
)
def listen(timeout, interval, scans):
    """Listen for SwitchBot meters and insert data into the temperature table.

    This command listens for SwitchBot meters and inserts the data into the temperature table.
    For each scan, wait for the specified timeout duration until all data is found and inserted once for each meter.
    The interval option specifies the time to wait between scans.


    Example:
    \b

        $ swb-meter listen
    \b

        $ swb-meter listen -t 10 -i 60 # Timeout 10 seconds, Interval 60 seconds
    """

    handler = DataHandler()

    m = get_meter_table()
    mac_address = [row["mac_address"] for row in m.rows_where(select="mac_address")]

    async def _run(scans=scans):
        while scans != 0:  # 無限ループでスキャンを繰り返す
            # a copy from global mac_address
            look_for = mac_address.copy()

            try:
                async with BleakScanner(
                    cb=dict(use_bdaddr=True)
                ) as scanner, async_timeout(timeout):
                    logger.info(f"Start scanning for {timeout} seconds")
                    async for bd, ad in scanner.advertisement_data():
                        logger.debug(f"Looking for {look_for}")

                        if bd.address not in look_for:
                            logger.debug(f"not found device {bd.address}")
                            continue

                        logger.info(f"Found device {bd.address}")
                        res = handler.on_data(ad)
                        if res:
                            look_for.remove(bd.address)

                        if not look_for:
                            logger.info("All devices found")
                            break

            except asyncio.TimeoutError:
                logger.info("Timeout")

            if scans:
                scans -= 1
                logger.info(f"Scans left: {scans}")

            await asyncio.sleep(interval)

    asyncio.run(_run())


@cli.command(name="streamlit")
def run_app():
    """Run the Streamlit app for visualizing the temperature data.

    Example:
    \b
    $ swb-meter app
    """
    # check if streamlit is installed
    # or suggest to install it via optional dependencies [streamlit]
    try:
        from streamlit.web import cli as stcli
    except ImportError:
        click.echo(
            "Streamlit is not installed. Install it via optional dependencies [streamlit]"
        )
        return

    import sys
    from pathlib import Path

    sys.argv = ["streamlit", "run", str(Path(__file__).parent / "streamlit.py")]
    sys.exit(stcli.main())



from datetime import timedelta

import pandas as pd
import plotly.express as px
import streamlit as st

from swb_meter.db import get_db

# SQLiteデータベースに接続
conn = get_db().conn

# データを読み込む
query = """
SELECT t.created_at, t.temperature, t.humidity,
       t.battery, t.rssi, t.mac_address, coalesce(m.alias, "no alias") as alias
FROM Temperature as t
LEFT JOIN Meter as m using(mac_address)
"""
df = pd.read_sql_query(query, conn)
df["created_at"] = pd.to_datetime(df["created_at"])  # タイムスタンプをdatetime型に変換


# 最新のデータの日時
latest_created_at = df["created_at"].max()

# タイムレンジの選択
time_range = st.sidebar.selectbox(
    "Select Time Range",
    options=["1 Hour", "1 Day", "1 Week", "1 Month", "1 Year"],
    index=0,
)


# タイムレンジに基づいてデータをフィルタリング
if time_range == "1 Hour":
    start_time = latest_created_at - timedelta(hours=1)
elif time_range == "1 Day":
    start_time = latest_created_at - timedelta(days=1)
elif time_range == "1 Week":
    start_time = latest_created_at - timedelta(weeks=1)
elif time_range == "1 Month":
    start_time = latest_created_at - timedelta(
        days=30
    )  # 精確な月の日数を計算する場合は、他の方法を検討してください
elif time_range == "1 Year":
    start_time = latest_created_at - timedelta(days=365)


st.title("SwitchBot Meter Data Dashboard")

# alias
aliases = df["alias"].unique()
selected_aliases = st.sidebar.multiselect("Select alias", aliases, default=aliases)


# フィルタリングされたデータフレーム
filtered_df = df[df["alias"].isin(selected_aliases)]


# aliasごとに最新のデータを表示
# 前回のデータを取得（2番目に新しいデータ）
latest_data = filtered_df.groupby("alias").last()
previous_data = filtered_df.groupby("alias").nth(-2).set_index("alias")

# aliasごとに最新のデータをメトリクスで表示（差分付き）
st.header("Latest Metrics")

metrics = ["temperature", "humidity"]  # , "バッテリー", "RSSI"]
for alias in selected_aliases:
    col1, col2, col3 = st.columns([1, 1, 1], gap="large", vertical_alignment="center")

    latest = latest_data.loc[alias]
    previous = previous_data.loc[alias]

    col1.subheader(alias)
    col1.text(latest["created_at"])

    col2.metric(
        label="温度",
        value=f"{latest['temperature']:.1f}℃",
        delta=f"{latest['temperature'] - previous['temperature']:.1f}℃",
    )
    col3.metric(
        label="湿度",
        value=f"{latest['humidity']:.1f}%",
        delta=f"{latest['humidity'] - previous['humidity']:.1f}%",
    )


st.header("Time Series")

labels = {"created_at": ""}

# 温度の時系列グラフ（2つのセンサーを1つのグラフに）
fig = px.line(
    filtered_df,
    x="created_at",
    y="temperature",
    color="alias",
    title="Temperature",
    labels=labels,
    markers=True,
    range_x=[start_time, latest_created_at],
)
st.plotly_chart(fig)

# 湿度の時系列グラフ（2つのセンサーを1つのグラフに）
fig = px.line(
    filtered_df,
    x="created_at",
    y="humidity",
    color="alias",
    title="Humidity",
    labels=labels,
)
st.plotly_chart(fig)

# バッテリー残量の時系列グラフ（2つのセンサーを1つのグラフに）
fig = px.line(
    filtered_df,
    x="created_at",
    y="battery",
    color="alias",
    title="Battery",
    labels=labels,
)
st.plotly_chart(fig)

# RSSIの時系列グラフ（2つのセンサーを1つのグラフに）
fig = px.line(
    filtered_df, x="created_at", y="rssi", color="alias", title="RSSI", labels=labels
)
st.plotly_chart(fig)

# 接続を閉じる
conn.close()



import os
from pathlib import Path

from sqlite_utils import Database


def get_db():
    db_path = Path(os.getenv("SWB_METER_DB_PATH", "./swb_meter.db"))

    if not db_path.exists():
        db_path.parent.mkdir(parents=True, exist_ok=True)
        db_path.touch()

    db = Database(db_path)

    return db


def get_temperature_table():
    db = get_db()
    t = db["Temperature"]

    return t


def get_meter_table():
    db = get_db()
    t = db["Meter"]

    return t



from dataclasses import dataclass
from enum import Enum

from swb_meter.logger import logger


@dataclass
class ParsedManufacturerData:
    company_code: str
    mac_address: str


@dataclass
class DeviceType:
    name: str
    mode: str
    code: str


@dataclass
class Group:
    A: bool
    B: bool
    C: bool
    D: bool


@dataclass
class TempAlert(Enum):
    NO_ALERT = 0
    LOW_TEMP_ALERT = 1
    HIGH_TEMP_ALERT = 2
    TEMP_ALERT = 3


@dataclass
class HumidAlert(Enum):
    NO_ALERT = 0
    LOW_HUMIDITY_ALERT = 1
    HIGH_HUMIDITY_ALERT = 2
    HUMIDITY_ALERT = 3


@dataclass
class TemperatureScale(Enum):
    CELSIUS = 0
    FAHRENHEIT = 1


@dataclass
class Temperature:
    value: float
    scale: str


@dataclass
class ParsedServiceData:
    uuid: str
    device_type: DeviceType
    group: Group
    battery: int
    temperature_alert: str
    humidity_alert: str
    temperature: Temperature
    humidity: int


@dataclass
class ParsedAdvertisementData:
    local_name: str
    rssi: int
    tx_power: int
    service_uuids: list
    manufacturer_data: ParsedManufacturerData
    service_data: ParsedServiceData


# SwitchBot Meter Service Data UUID
SERVICE_DATA_UUID = "0000fd3d-0000-1000-8000-00805f9b34fb"

# SwitchBot Company Code
COMPANY_CODE = "0x969"


def get_device_type(type_code):
    device_types = {
        0x48: ("SwitchBot Bot (WoHand)", None),
        0x42: ("WoButton", None),
        0x4C: ("SwitchBot Hub (WoLink)", "Add Mode"),
        0x6C: ("SwitchBot Hub (WoLink)", "Normal Mode"),
        0x50: ("SwitchBot Hub Plus (WoLink Plus)", "Add Mode"),
        0x70: ("SwitchBot Hub Plus (WoLink Plus)", "Normal Mode"),
        0x46: ("SwitchBot Fan (WoFan)", "Add Mode"),
        0x66: ("SwitchBot Fan (WoFan)", "Normal Mode"),
        0x74: ("SwitchBot MeterTH (WoSensorTH)", "Add Mode"),
        0x54: ("SwitchBot MeterTH (WoSensorTH)", "Normal Mode"),
        0x4D: ("SwitchBot Mini (HubMini)", "Add Mode"),
        0x6D: ("SwitchBot Mini (HubMini)", "Normal Mode"),
    }
    return device_types.get(type_code, ("Unknown Device", None))


def parse_advertisement_data(
    advertisement_data,
    service_data_uuid=SERVICE_DATA_UUID,
):
    manufacturer_data = None

    # manufacturer_dataは, {会社コード: データ}の辞書
    # eg: {76: b'\x16\x08\x00\x98\x91\xb4j\xd8\x11\xfc'}
    for company_code_bin, data in advertisement_data.manufacturer_data.items():
        # int as byte to hex string
        company_code = hex(company_code_bin)

        if company_code != COMPANY_CODE:
            logger.debug(f"Company code does not match: {company_code}")
            return

        # switchbot meterのmacアドレスを取得
        mac_address = ":".join([f"{b:02X}" for b in data[:6]])

        manufacturer_data = ParsedManufacturerData(
            company_code=company_code, mac_address=mac_address
        )

    service_data = None

    for uuid, data in advertisement_data.service_data.items():
        # Service Data UUIDが一致しない場合はスキップ
        if not uuid or uuid != service_data_uuid:
            return

        # Byte 0: Device Type
        device_type_code = data[0] & 0x7F
        device_name, mode = get_device_type(device_type_code)
        device_type = DeviceType(
            name=device_name, mode=mode, code=hex(device_type_code)
        )

        # Byte 1: Status
        status = data[1]
        group = Group(
            A=bool(status & 0x01),
            B=bool(status & 0x02),
            C=bool(status & 0x04),
            D=bool(status & 0x08),
        )

        # Byte 2: Battery
        battery = data[2] & 0x7F

        # Byte 3: Temperature Alert, Humidity Alert, Temperature Decimal
        temp_alert = (data[3] >> 6) & 0x03
        humid_alert = (data[3] >> 4) & 0x03
        temp_decimal = data[3] & 0x0F

        temperature_alert = TempAlert(temp_alert).name
        humidity_alert = HumidAlert(humid_alert).name

        # Byte 4: Temperature
        temp_positive = bool(data[4] & 0x80)
        temp_integer = data[4] & 0x7F

        # Byte 5: Temperature Scale and Humidity
        humidity = data[5] & 0x7F

        # Calculate full temperature
        temperature = temp_integer + (temp_decimal / 10)
        if not temp_positive:
            temperature = -temperature

        temperature = Temperature(
            value=temperature, scale=TemperatureScale(data[5] & 0x80).name
        )

        service_data = ParsedServiceData(
            uuid=uuid,
            device_type=device_type,
            group=group,
            battery=battery,
            temperature_alert=temperature_alert,
            humidity_alert=humidity_alert,
            temperature=temperature,
            humidity=humidity,
        )

    # 何も取得できなかった場合はNoneを返す
    if not manufacturer_data or not service_data:
        logger.debug("No data in advertisement data")
        return

    parsed_data = ParsedAdvertisementData(
        local_name=advertisement_data.local_name,
        rssi=advertisement_data.rssi,
        tx_power=advertisement_data.tx_power,
        service_uuids=advertisement_data.service_uuids,
        manufacturer_data=manufacturer_data,
        service_data=service_data,
    )

    return parsed_data
